(*---------------------------------------------------------------------------*
  INTERFACE  iom_layer.mli

  Copyright (c) 2006, James H. Woodyatt
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Stream commutation layers.  *)

(** {6 Overview}

    Classes and functions for commutating I/O event stream gadgets are defined.
    Both simplex (input and output) and duplex streams are supported.  In
    addition, convenience functions are provided for creating commutationg
    layer gadgets from initial state objects.
*)

(** {5 Transform Class} *)

(** Use [inherit transform rx tx] to derive a substate object that can consume
    events of the type carried by [rx] and produce events of the type carried
    by [tx].
*)
class virtual ['x, 'y] transform:
    'x #Iom_gadget.rx -> 'y #Iom_gadget.tx ->
    object('self)
        val consume_: bool  (** [true] if ready to consume events. *)
        val produce_: bool  (** [true] if unblocked for producing events. *)

        (** The result of [self#push event] is applied in the [consume] method
            to obtain the substate object after [event] is processed.
        *)
        method virtual push: 'x -> 'self Iom_gadget.t

        (** The result of [self#shift] is evaluated to obtain a list of events
            to produce and the optional substate to return in the [produce]
            method.
        *)
        method virtual shift: ('y list * 'self option) Cf_exnopt.t Iom_gadget.t

        (** If the stream is ready to consume events, then [self#consume f]
            guards on the receiver, dispatching received events through
            [self#push] and applying [f] to the resulting substate object.  If
            [push] raises an exception, then [f] is applied to the exception
            instead.  If the stream is not ready to consume events, then the
            receiver is not guarded.
        *)
        method consume:
            ('self Cf_exnopt.t -> unit Iom_gadget.t) -> unit Iom_gadget.guard

        (** If the stream is unblocked for producing events, then
            [self#produce] returns an optional substate indicating that further
            processing is expected, returns [None] to indicate that processing
            is finished, or returns an exception.  If the stream is not ready
            for producing events, then the current substate is returned with no
            other effect.
        *)
        method produce: 'self option Cf_exnopt.t Iom_gadget.t

        (** Returns a copy of the substate with [consume_ = false]. *)
        method wait: 'self Iom_gadget.t

        (** Returns a copy of the substate with [consume_ = true]. *)
        method ready: 'self Iom_gadget.t

        (** Returns a copy of the substate with [produce_ = true]. *)
        method unblock: 'self Iom_gadget.t

        (** Returns a copy of the substate with [produce_ = false]. *)
        method block: 'self Iom_gadget.t
    end

(** {5 Process Classes} *)

(** The class type describing the common features of both input and output
    simplex stream states.
*)
class type ['xy, 'c, 'n] simplex = object('self)
    inherit ['c, 'n] Iom_stream.process
    constraint 'xy = ('x, 'y) #transform
    
    val transform_: 'xy (** A delegated stream transform object. *)

    (** A delegator for the [wait] method on the transform object. *)
    method private wait: unit Iom_gadget.t

    (** A delegator for the [ready] method on the transform object. *)
    method private ready: unit Iom_gadget.t

    (** A delegator for the [unblock] method on the transform object. *)
    method private unblock: unit Iom_gadget.t

    (** A delegator for the [block] method on the transform object. *)
    method private block: unit Iom_gadget.t
end

(** The class of stream gadget state for input stream commutators.  Control
    events are constrained to include flow control signals.  Use [inherit
    \['xy, 'c, 'n\] isimplex xy fix pad] to derive a state class object that
    delegates stream event flow control to the transform [xy], receives
    upstream control and sends upstream notifications on [pad], and also sends
    downstream control and receives downstream notifications on [fix].
*)
class ['xy, 'c, 'n] isimplex:
    'xy -> ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    ['xy, [> Iom_stream.flowcontrol ] as 'c, 'n] simplex

(** The class of stream gadget state for output stream commutators.
    Notification events are constrained to include flow control signals.  Use
    [inherit \['xy, 'c, 'n\] osimplex xy fix pad] to derive a state class
    object that delegates stream event flow control to the transform [xy],
    receives upstream control and sends upstream notifications on [pad], and
    also sends downstream control and receives downstream notifications on
    [fix].
*)
class ['xy, 'c, 'n] osimplex:
    'xy -> ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    ['xy, 'c, [> Iom_stream.flownotify ] as 'n] simplex

(** The class of stream gadget state for full duplex stream commutators.
    Both control and notification events are constrained to include flow
    control signals.  Use [inherit \['ixy, 'oxy, 'c, 'n\] duplex ixy oxy fix
    pad] to derive a state class object that delegates stream event flow
    control to the input transform [ixy] and the output transform [oxy],
    receives upstream control and sends upstream notifications on [pad], and
    also sends downstream control and receives downstream notifications on
    [fix].
*)
class ['ixy, 'oyx, 'c, 'n] duplex:
    'ixy -> 'oyx ->
    ('n #Iom_gadget.rx * 'c #Iom_gadget.tx) ->
    ('c #Iom_gadget.rx * 'n #Iom_gadget.tx) ->
    object('self)
        inherit ['c, 'n] Iom_stream.notify
        inherit ['c, 'n] Iom_stream.control
        inherit Iom_gadget.next
        inherit Iom_gadget.start
        constraint 'c = [> Iom_stream.flowcontrol ]
        constraint 'n = [> Iom_stream.flownotify ]
        constraint 'ixy = ('ix, 'iy) #transform
        constraint 'oyx = ('oy, 'ox) #transform
        
        (** A delegated input stream transform object. *)
        val itransform_: 'ixy option

        (** A delegated output stream transform object. *)
        val otransform_: 'oyx option

        (** A delegator for [iwait] on the input transform object. *)
        method private iwait: unit Iom_gadget.t

        (** A delegator for [iready] on the input transform object. *)
        method private iready: unit Iom_gadget.t

        (** A delegator for [iunblock] on the input transform object. *)
        method private iunblock: unit Iom_gadget.t

        (** A delegator for [iblock] on the input transform object. *)
        method private iblock: unit Iom_gadget.t

        (** A delegator for [owait] on the output transform object. *)
        method private owait: unit Iom_gadget.t

        (** A delegator for [oready] on the output transform object. *)
        method private oready: unit Iom_gadget.t

        (** A delegator for [ounblock] on the output transform object. *)
        method private ounblock: unit Iom_gadget.t

        (** A delegator for [oblock] on the output transform object. *)
        method private oblock: unit Iom_gadget.t
    end

(** Use [ingest fix cons] to contruct a harness of wires with a new pad and
    fix for input streams on its respective ends, apply the appropriate wire
    ends (a control event transmitter, an input receiver and an output
    transmitter) to [cons], obtain a stream transform object, and use it to
    construct an [isimplex] object and start the input stream gadget.  The new
    fix is returned.
*)
val ingest:
  ('x, [> Iom_stream.flowcontrol ] as 'c, [> Iom_stream.failed ] as 'n)
    Iom_stream.ifix ->
  ('c Iom_gadget.tx -> 'x Iom_gadget.rx -> 'y Iom_gadget.tx ->
    ('x, 'y) #transform) ->
  ('y, 'c, 'n) Iom_stream.ifix Iom_gadget.t

(** Use [render fix cons] to contruct a harness of wires with a new pad and
    fix for output streams on its respective ends, apply the appropriate wire
    ends (a notification event transmitter, an input receiver and an output
    transmitter) to [cons], obtain a stream transform object, and use it to
    construct an [osimplex] object and start the output stream gadget.  The new
    fix is returned.
*)
val render:
  ('y, [> Iom_stream.stop ] as 'c, [> Iom_stream.flownotify ] as 'n)
    Iom_stream.ofix ->
  ('n Iom_gadget.tx -> 'x Iom_gadget.rx -> 'y Iom_gadget.tx ->
    ('x, 'y) #transform) ->
  ('x, 'c, 'n) Iom_stream.ofix Iom_gadget.t

(** Use [duplex fix icons ocons] to contruct a harness of wires with a new
    pad and fix for output streams on its respective ends, apply the
    appropriate wire ends to [icons] and [ocons], obtain a pair of stream
    transform objects, and use them to construct an [duplex] object and start
    the output stream gadget.  The new fix is returned.
*)
val duplex:
    ('xi, 'xo, [> Iom_stream.flowcontrol ] as 'c,
        [> Iom_stream.flownotify ] as 'n) Iom_stream.iofix ->
    ('c Iom_gadget.tx -> 'xi Iom_gadget.rx -> 'yi Iom_gadget.tx ->
        ('xi, 'yi) #transform) ->
    ('n Iom_gadget.tx -> 'yo Iom_gadget.rx -> 'xo Iom_gadget.tx ->
        ('yo, 'xo) #transform) ->
    ('yi, 'yo, 'c, 'n) Iom_stream.iofix Iom_gadget.t

(*--- End of File [ iom_layer.mli ] ---*)
